__version__ = "$Id: Grammar.py,v 0.1 1997/12/08 09:22:15 scott Exp $"

#
# Error checking.  These errors correspond the the same errors in
# Yacc/Bison
#
class ConflictError:

    def __init__(self, state_index, terminal_index, old, new):
	self.indices = (state_index, terminal_index)
	self.old = old
	self.new = new


class ShiftShiftError(ConflictError):
    pass

class ShiftReduceError(ConflictError):
    pass

class ReduceReduceError(ConflictError):
    pass


#
# Production -- a Grammar is really just a list of productions.
# The expected structure is a symbol for the LHS and a list of 
# symbols or symbols for the RHS.
#
class Production:

    def __init__(self, LHS, RHS):
	self.LHS = LHS
	self.RHS = RHS
	self.func = None
	self.funcname = ""

    #
    # .setfunc(<callable>)  --used for the dynamic production 
    # of a parseengine directly from Grammar.mkengine(), instead of tables 
    # saved to a file.
    #
    def setfunc(self, func):
	self.func = func
	
    #
    # .setfuncname("") -- used by Grammar.writefile to produce
    # prodinfo table that.  .setfunc associates a function value
    # with the production for runtime, on the fly productions
    # of parsing engine from Grammar.
    #
    def setfuncname(self, name):
	self.funcname = name

    def __len__(self):
	return len(self.RHS)

    def __repr__(self):
	return self.LHS + " -> " + `self.RHS`

    def items(self):
	return range(len(self.RHS) + 1)

#
# Perhaps this should be called 'LR1Grammar' instead...
# Provides methods for producing the actiontable, the gototable, and the 
# prodinfo table.  Using these functions, it can produce a python source code file
# with these tables or a parsing engine.
#
class Grammar:

    EPS = "<EPS>"
    EOF = "<EOF>"
    DummyLA = -1

    def __init__(self, prods, toks=None):
	self.productions = prods[:]
	if toks:
	    self.tokens = toks
	else:
	    self.tokens = []
	self.terms = self.terminals()
	self.nonterms = self.nonterminals()

    def terminals(self):
	t = []
	for p in self.productions:
	    for s in p.RHS:
		if s not in t:
		    t.append(s)
	for nt in self.nonterminals():
	    if nt in t:
		t.remove(nt)
	t.sort()
	return t

    def nonterminals(self):
	nt = []
	for p in self.productions:
	    if p.LHS not in nt:
		nt.append(p.LHS)
	return nt


    def expansions(self, LHS):
	e = []
	for p in self.productions:
	    if p.LHS == LHS:
		e.append(p.RHS)
	return e

    def LHSproductions(self, lhs):
	res = []
	for p in self.productions:
	    if p.LHS == lhs:
		res.append(p)
	return res

    def showexp(self, LHS):
	import string
	s = LHS + " -> "
	w = len(s)
	exps = self.expansions(LHS)
	for e in exps[:-1]:
	    for ei in range(len(e)):
		if type(e[ei]) == type(0):
		    if self.tokens:
			e[ei] = self.tokens[e[ei]]
		    else:
			e[ei] = str(e[ei])
	    s = s + string.join(e, " ") + " |\n" + " " * w
	e = exps[-1]
	for ei in range(len(e)):
	    if type(e[ei]) == type(0):
		if self.tokens:
		    e[ei] = self.tokens[e[ei]]
		else:
		    e[ei] = str(e[ei])	    
	s = s + string.join(e, " ") + ";"
	return s


    def __repr__(self):
	s = ""
	for nt in self.nonterminals():
	    s = s + self.showexp(nt) + "\n\n"
	return s


    def augment(self):
	lhss = map(lambda x: x.LHS, self.productions)
	newsym = self.productions[0].LHS
	while 1:
	    newsym = newsym + "'"
	    if newsym not in lhss:
		break
	self.productions.insert(0, Production(newsym, 
					      [self.productions[0].LHS]))

				

    def first(self, sym, banned=None):
	if not banned:
	    banned = []
	if sym in self.terms or sym == Grammar.DummyLA:
	    return [sym]
	symprods = self.LHSproductions(sym)
	res = []
	banned.append(sym)
	for sp in symprods:
	    if sp.RHS == [Grammar.EPS] and Grammar.EPS not in res:
		res.append(Grammar.EPS)
	    else:
		allhaveeps = 1
		for s in sp.RHS:
		    if s in banned:
			continue
		    subfirst = self.first(s, banned)
		    res = res + subfirst
		    if Grammar.EPS not in subfirst:
			allhaveeps = 0
			break
		if allhaveeps and Grammar.EPS not in res:
		    res.append(Grammar.EPS)
	return res


    def firstofstring(self, gs_list):
	res = []
	allhaveeps = 1
	for x in range(len(gs_list)):
	    tmp = self.first(gs_list[x])
	    if Grammar.EPS in tmp:
		tmp.remove(Grammar.EPS)
	    else:
		allhaveeps = 0
	    res = res + tmp
	if allhaveeps:
	    res.append(Grammar.EPS)
	return res


    def follow(self):
	eof = Grammar.EOF
	follow = {}
	startsym = self.productions[0].LHS
	follow[startsym] = [eof]
	nts = self.nonterms
	for p in self.productions:
	    cutoff = range(len(p.RHS))
	    cutoff.reverse()
	    for c in cutoff[:-1]:  # all but the first of the RHS elements
		f = self.first(p.RHS[c])
		if Grammar.EPS in f:
		    f.remove(Grammar.EPS)
		if follow.has_key(p.RHS[c - 1]):
		    if p.RHS[c -1] in nts:
			follow[p.RHS[c -1]] = follow[p.RHS[c - 1]] + f[:]
		else:
		    if p.RHS[c -1] in nts:
			follow[p.RHS[c - 1]] = f[:]
 	for p in self.productions:
 	    cutoff = range(len(p.RHS))
 	    cutoff.reverse()
	    if p.RHS[-1] in nts:
		if follow.has_key(p.LHS):
		    add = follow[p.LHS]
		else:
		    add = []

		if follow.has_key(p.RHS[-1]):
		    follow[p.RHS[-1]] = follow[p.RHS[-1]] + add
		else:
		    follow[p.RHS[-1]] = add
 	    for c in cutoff[:-1]:
 		f = self.first(p.RHS[c])
 		if Grammar.EPS in f:
 		    if follow.has_key(p.LHS):
 			add = follow[p.LHS]
 		    else:
 			add = []
 		    if follow.has_key(p.RHS[c-1]):
 			follow[p.RHS[c-1]] = follow[p.RHS[c-1]] + add
 		    elif add:
 			follow[p.RHS[c - 1]] = add
	for k in follow.keys():
	    d = {}
	    for i in follow[k]:
		d[i] = 1
	    follow[k] = d.keys()
	return follow


    def closure(self, items):
	res = items[:]
	while 1:
	    more = []
	    for (prodind, rhsind), term in res:
		if rhsind == len(self.productions[prodind].RHS):
		    continue
		for p in self.LHSproductions(self.productions[prodind].RHS[rhsind]):
		    try:
			newpart = self.productions[prodind].RHS[rhsind + 1]
		    except IndexError:
			newpart = Grammar.EPS
		    stringofsyms = [newpart, term]
		    for t in self.firstofstring(stringofsyms):
			if ((self.productions.index(p), 0), t) not in res:
			    more.append(((self.productions.index(p), 0), t))
		    if term == Grammar.EOF and newpart == Grammar.EPS:
			if ((self.productions.index(p), 0), Grammar.EOF) not in res:
			    more.append(((self.productions.index(p), 0), Grammar.EOF))
	    if more:
		res = res + more
	    else:
		break
	return res


    def goto(self, items, sym):
	itemset = []
	for (prodind, rhsind), term in items:
	    try:
		if self.productions[prodind].RHS[rhsind] == sym and ((prodind, rhsind+1), term) not in itemset:
		    itemset.append( ((prodind, rhsind +1), term))
	    except IndexError:  
		pass
	return self.closure(itemset)

    #
    # LR1items -- this operation is extremely expensive.  Try not to call it
    # more than once per production of tables
    #
    def LR1items(self, save=0):
	res = [self.closure([((0,0), Grammar.EOF)])]
	syms = self.nonterms + self.terms
	while 1:
	    more = []
	    for itemset in res:
		for s in syms:
		    tmp = self.goto(itemset, s)
		    if tmp and tmp not in res and tmp not in more:
			more.append(tmp)
	    if not more:
		if save:
		    self.LR1items = res
		return res
	    else:
		res = res + more
	
    def actiontable(self, items=None):
	if not items:
	    items = self.LR1items()
	res = []
	state_i = 0
	terms = self.terms[:]
	terms.append(Grammar.EOF)
	errentry = ("", -1)
	for state in items:
	    list = [errentry] * len(terms)
	    res.append(list)
	    for (prodind, rhsind), term in state:
		if (rhsind ) == len(self.productions[prodind].RHS):
		    if prodind != 0:
			new = ("r", prodind)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
		    else:
			new = ("a", -1)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
		elif self.productions[prodind].RHS[rhsind] in terms:
		    tmp = self.goto(state, self.productions[prodind].RHS[rhsind])
		    if tmp in items:
			inner_i = terms.index(self.productions[prodind].RHS[rhsind])
			new =  ("s", items.index(tmp))
			old = res[state_i][inner_i]
			if old != errentry and old != new:
			    print "Conflict[%d, %d]:" % (state_i, inner_i), old, "->", new
			res[state_i][inner_i] = new
	    state_i = state_i + 1
	return res


    def gototable(self, items=None):
	if not items:
	    items = self.LR1items()
	res = []
	state_i = 0
	nonterms = self.nonterms
#	nonterms.remove(self.productions[0].LHS)
	err = None
	for state in items:
	    list = [err] * len(nonterms)
	    res.append(list)
	    nonterm_i = 0
	    for nt in nonterms:
		goto = self.goto(state, nt)
		if goto in items:
		    res[state_i][nonterm_i] = items.index(goto)
		nonterm_i = nonterm_i + 1
	    state_i = state_i + 1
	return res


    def setfunc(self, prodind, func):
	self.productions[prodind].setfunc(func)

    
    def setfuncname(self, prodind, funcname):
	self.productions[prodind].setfuncname(funcname)

    def default_prodfunc(self):
	return lambda *args: args[0]

    def prodinfotable(self):
	res = []
	nonterms = self.nonterms
	for p in self.productions:
	    lhsind = nonterms.index(p.LHS)
	    func = p.func
	    if not func:
		func = self.default_prodfunc()
	    plen = len(p.RHS)
	    res.append((plen, func, lhsind))
	return res



    def mkengine(self, inbufchunksize=None, stackchunksize=None):
	import PyLRengine
	self.augment()
	items = self.LR1items()
	at = self.actiontable(items)
	gt = self.gototable(items)
	self.productions = self.productions[1:]  # unaugment
	pi = self.prodinfotable()
	if not inbufchunksize:
	    inbufchunksize = 50
	if not stackchunksize:
	    stackchunksize = 100
	e = PyLRengine.NewEngine(pi, at, gt, inbufchunksize, stackchunksize)
	return e


    def writefile(self, filename):
	import PyLRtabletemplate
	template = PyLRtabletemplate.__doc__
	vals = {}
	import time
	vals["date"] = time.ctime(time.time())
	vals["filename"] = filename
	if not hasattr(self, "extrasource"):
	    vals["extrasource"] = ""
	else:
	    vals["extrasource"] = self.extrasource
	vals["grammar"] = `self`
	self.augment()
	lr1items = self.LR1items()
	actiontable = self.actiontable(lr1items)
	actiontable_s = "[\n\t"
	for l in actiontable:
	    actiontable_s = "%s%s,\n\t" % (actiontable_s, `l`)
	vals["actiontable"] = actiontable_s[:-3] + "\n]\n\n"
	gototable = self.gototable(lr1items)
	gototable_s = "[\n\t"
	for l in gototable:
	    gototable_s = "%s%s,\n\t" % (gototable_s, `l`)
	vals["gototable"] = gototable_s[:-3] + "\n]\n\n"
	self.productions = self.productions[1:]  # unaugment
	pi = self.prodinfotable()
	pi_s = "[\n\t"
	for l, f, e in pi:
	    pi_s = "%s(%d, _fdummy, %d),\n\t" % (pi_s, l,e )
	vals["prodinfo"] = pi_s + "]\n\n"
	fp = open(filename, "w")
	fp.write(template % vals)
	fp.close()


class LALRGrammar(Grammar):



    def __init__(self, prods, toks=None):
	Grammar.__init__(self, prods, toks)
	self.LALRitems = []


    def LALRclosure(self, itemset):
	res = itemset[:]
	while 1:
	    more = []
	    for (prodind, rhsind) in res:
		if rhsind == len(self.productions[prodind].RHS):
		    continue
		for p in self.LHSproductions(self.productions[prodind].RHS[rhsind]):
		    newprodind = self.productions.index(p)
		    if (newprodind, 0) not in res:
			more.append((newprodind, 0))
	    if more:
		res = res + more
	    else:
		break
	return res


    def goto(self, itemset, sym):
	items = []
	for (prodind, rhsind) in itemset:
	    try:
		if (self.productions[prodind].RHS[rhsind] == sym and 
		    (prodind, rhsind+1) not in items):
		    items.append((prodind, rhsind +1))
	    except IndexError:  
		pass
	res = self.LALRclosure(items)
	res.sort()
	return res

    def LR0items(self):
	res = [self.LALRclosure([(0,0)])]
	syms = self.nonterms + self.terms
	while 1:
	    more = []
	    for iset in res:
		for s in syms:
		    g = self.goto(iset, s)
		    if g and g not in res:
			more.append(g)
	    if not more:
		break
	    else:
		res = res + more
    	for li in range(len(res)):
    	    res[li].sort()
	self.soi = res
	return res


    def kernels(self, items):
	res = []
	for (prodind, rhsind) in items:
	    if (rhsind != 0) or (prodind == 0 and rhsind == 0):
		res.append((prodind, rhsind))
	return res


    def lookaheads(self, itemset):
	setsofitems = self.soi
	state_i = setsofitems.index(itemset)
	spontaneous = []
	propagates = {}
	kernels = self.kernels(itemset)
	for (kpi, kri) in kernels:
	    C = self.closure([((kpi, kri), Grammar.DummyLA)])
	    for (cpi, cri), t in C:
		if (cri) >= len(self.productions[cpi].RHS):
		    continue
		try:
		    newstate = setsofitems.index(self.goto(itemset, self.productions[cpi].RHS[cri]))
		except ValueError:
		    continue
		if t != Grammar.DummyLA:
		    spontaneous.append((newstate, (cpi, cri+1), t))
		else:
 		    if propagates.has_key((kpi, kri)):
			propagates[(kpi, kri)].append((newstate, (cpi, cri+1)))
		    else:
			propagates[(kpi, kri)]=[(newstate, (cpi, cri+1))]
	return spontaneous, propagates
		

    def initLALR1items(self): 
	setsofitems = self.LR0items()
	props = {}
	la_table = []
	for x in range(len(setsofitems)):
	    la_table.append([])
	    for y in range(len(setsofitems[x])):
		la_table[x].append([])
	la_table[0][0] = [Grammar.EOF]
	state_i = 0
	for itemset in setsofitems:
            sp, pr = self.lookaheads(itemset)
	    for ns, (pi, ri), t in sp:
		inner = setsofitems[ns].index((pi, ri))
		la_table[ns][inner].append(t)
	    props[state_i] = pr
	    state_i = state_i + 1
	return la_table, props
    

    def LALR1items(self):
	la_table, props = self.initLALR1items()
	soi = self.soi
	while 1:
	    added_la = 0
	    state_i = 0
	    for state in la_table:
		ii = 0
		for propterms in state:
		    if not propterms:
			ii = ii + 1
			continue
		    item = soi[state_i][ii]
		    ii = ii + 1
		    try:
			proplist = props[state_i][item]
		    except KeyError:
			continue
		    for pstate, pitem in proplist:
			inner = soi[pstate].index(pitem)
			for pt in propterms:
			    if pt not in la_table[pstate][inner]:
				added_la = 1
				la_table[pstate][inner].append(pt)
		state_i = state_i + 1
	    if not added_la:
		break
	#
	# this section just reorganizes the above data
	# to the state it's used in later...
	# 
	res = []
	state_i = 0
	for state in soi:
	    item_i = 0
	    inner = []
	    for item in state:
		for term in la_table[state_i][item_i]:
		    if (item, term) not in inner:
			inner.append((item, term))
		item_i = item_i + 1
	    inner.sort()
	    res.append(inner)
	    state_i = state_i + 1
	self.LALRitems = res
	return res


    def actiontable(self):
	items = self.LALRitems
	res = []
	state_i = 0
	terms = self.terms[:]
	terms.append(Grammar.EOF)
	errentry = ("", -1)
	for state in items:
	    list = [errentry] * len(terms)
	    res.append(list)
	    for (prodind, rhsind), term in state:
		if (rhsind ) == len(self.productions[prodind].RHS):
		    if prodind != 0:
			new = ("r", prodind)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
		    else:
			new = ("a", -1)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
	    state_i = state_i + 1
	res = self.addshifts(res)
	return res


    def addshifts(self, res):
	state_i = 0
	for state in self.soi:
	    for pi, ri in state:
		try:
		    if self.productions[pi].RHS[ri] in self.terms:
			gt = self.goto(state, self.productions[pi].RHS[ri])
			res[state_i][self.terms.index(self.productions[pi].RHS[ri])] = ("s", self.soi.index(gt))
		except ValueError:
		    pass
		except IndexError:
		    pass
	    state_i = state_i + 1
	return res
	    

    def gototable(self):
	items = self.soi
	res = []
	state_i = 0
	nonterms = self.nonterms
	err = None
	for state in items:
	    list = [err] * len(nonterms)
	    res.append(list)
	    nonterm_i = 0
	    for nt in nonterms:
		goto = self.goto(state, nt)
		if goto in items:
		    res[state_i][nonterm_i] = items.index(goto)
		nonterm_i = nonterm_i + 1
	    state_i = state_i + 1
	return res


    def mkengine(self, inbufchunksize=None, stackchunksize=None):
	import PyLRengine
	self.augment()
	self.LALR1items()
	at = self.actiontable()
	gt = self.gototable()
	self.productions = self.productions[1:]  # unaugment
	pi = self.prodinfotable()
	if not inbufchunksize:
	    inbufchunksize = 50
	if not stackchunksize:
	    stackchunksize = 100
	e = PyLRengine.NewEngine(pi, at, gt, inbufchunksize, stackchunksize)
	return e


    def writefile(self, filename, parsername="MyParser", lexerinit = "PyLR.Lexer.Lexer()"):
	self.augment()
	self.LALR1items()
	at = self.actiontable()
	gt = self.gototable()
	self.productions = self.productions[1:]
	pi = self.prodinfotable()
	import parsertemplate
	template = parsertemplate.__doc__
	vals = {"parsername": parsername, "lexerinit": lexerinit}
	import time
	vals["date"] = time.ctime(time.time())
	vals["filename"] = filename
	if not hasattr(self, "extrasource"):
	    vals["extrasource"] = ""
	else:
	    vals["extrasource"] = self.extrasource
	vals["grammar"] = `self`
	actiontable_s = "[\n\t"
	for l in at:
	    actiontable_s = "%s%s,\n\t" % (actiontable_s, `l`)
	vals["actiontable"] = actiontable_s[:-3] + "\n]\n\n"
	gototable_s = "[\n\t"
	for l in gt:
	    gototable_s = "%s%s,\n\t" % (gototable_s, `l`)
	vals["gototable"] = gototable_s[:-3] + "\n]\n\n"
	pi_s = "[\n\t"
	pii = 0
	vals["symbols"] = `self.tokens + ["EOF"]`
	for l, f, e in pi:
	    pi_s = "%s(%d, '%s', %d),%s# %s\n\t" % (pi_s, 
						    l, 
						    self.productions[pii].funcname, 
						    e, 
						    " " * (18 - len(self.productions[pii].funcname)),
						    `self.productions[pii]` )
	    
	    pii = pii + 1
	vals["prodinfo"] = pi_s + "]\n\n"
	fp = open(filename, "w")
	fp.write(template % vals)
	fp.close()

	

def _test():
    # the token interface will be defined in 
    # a wrapper module and in the lexer.
    # this is the format the information will be stored in
    #
    # first, define the tokens
    #
    (id, plus, times, lparen, rparen, eof) = range(6)
    #
    # then, define how you want them to appear
    # in the documentation of the output file
    #
    toks = ["id", "+", "*", "(", ")"]

    #
    # define the productions (LHS=left hand side, RHS=right hand side)
    #
    prods = map(lambda x: Production(x[0], x[1]), [("E", ["E", plus, "T"]),
						   ("E", ["T"]), 
						   ("T", ["T", times, "F"]), 
						   ("T", ["F"]),
						   ("F", [lparen, "E", rparen]),
						   ("F", [id])])
    global g
    g = LALRGrammar(prods, toks)
    #

    #
    # define functions for the parser to use
    #
    def fadd(*args):
	print "adding %d with %d" % (args[0], args[2])
	return args[0] + args[2]

    def fdummy(*args):
	print "calling fdummy with args %s" % `args`
	return args[0]

    def ftimes(*args):
	print "multiplying %d with %d" % (args[0], args[2])
	return args[0] * args[2]

    def fparens(*args):
	print "handling parens, returning whats in between"
	return args[1]

    
    #
    # register the functions
    #
    for i in range(len(g.productions)):
	if len(g.productions[i].RHS) == 1: 
	    g.setfunc(i, fdummy)
    g.setfunc(0, fadd)
    g.setfunc(2, ftimes)
    g.setfunc(4, fparens)
    #
    # produce the parser
    #
    print "generating parser"
    e = g.mkengine()
    print "parser generated"
    #
    # this is the input as would be returned by the Lexer 
    #
    input = [(lparen, "("), (id, 3), (plus, "+"),  
	     (lparen, "("), (id, 4), (times, "*"), (id, 2),  (rparen, ")"), 
	     (rparen, ")"), (times, "*"), 
	     (id, 2), (times, "*"), (id, 5) ]
    #
    # the parser keeps parsing until either an error occurs or an accept state occurs.  There
    # is no special EOF symbol for it to stop taking input yet, so for the moment you have to tag
    # a bunch of them on the end of the input :(
    #
    eofs = [(eof, "")] * 15
    input = input + eofs
    for token, val in input:
	e.parse(token, val)
    


    


   



    





